#include "Camera.h"
#include <iostream>
#include <glm/gtx/spline.hpp> // For glm::catmullRom

Camera::Camera() : pathTime(0.0f) {}

void Camera::setControlPoints(const std::vector<glm::vec3>& points) {
	controlPoints = points;
	totalTime = 2.0f; // time to complete the movement
	pathTime = 0.0f; // Reset path time when new points are set
}



glm::vec3 Camera::getPositionOnSpline(float t) {
	if (controlPoints.size() < 4) {
		return position; // Return the current position if there aren't enough points
	}

	// Ensure t is wrapped around the path
	int numSegments = controlPoints.size() - 3; // Number of spline segments
	int currentSegment = static_cast<int>(t) % numSegments;
	float segmentT = t - static_cast<int>(t); // Fractional part for interpolation

	// Retrieve the control points for the current segment
	glm::vec3 p0 = controlPoints[currentSegment];
	glm::vec3 p1 = controlPoints[currentSegment + 1];
	glm::vec3 p2 = controlPoints[currentSegment + 2];
	glm::vec3 p3 = controlPoints[currentSegment + 3];

	// Interpolate position using Catmull-Rom spline
	return glm::catmullRom(p0, p1, p2, p3, segmentT);
}

bool Camera::followPathLookAhead(GLfloat deltaTime) {
	if (controlPoints.size() < 4) {
		throw std::runtime_error("Fewer than 4 ctr points in followPathLookAhead!");
		return true;
	}

	float speed = deltaTime / totalTime;
	pathTime += speed;
		
	// Check if the end is reached (small offset because we look ahead)
	if (pathTime >= 1.0f - 0.0015) {
		pathTime = 1.0f;
		return true;
	}

	// Calculate the new position on the spline
	glm::vec3 newPosition = getPositionOnSpline(pathTime);

	// Update camera's position and direction
	position = newPosition;

	// Optional: Look ahead to orient the camera along the path
	glm::vec3 nextPosition = getPositionOnSpline(pathTime + 0.001f); // Small offset to look ahead
	front = glm::normalize(nextPosition - position);
	right = glm::normalize(glm::cross(front, worldUp));
	up = glm::normalize(glm::cross(right, front));
	return (pathTime >= 1.0f);
}

bool Camera::followPathLookDirection(GLfloat deltaTime, const glm::vec3* direction = nullptr) {
	if (controlPoints.size() < 4) {
		throw std::runtime_error("Fewer than 4 ctr points in followPathLookDirection!");
		return true;
	}

	float speed = deltaTime / totalTime;
	pathTime += speed;

	if (pathTime >= 1.0f) {
		pathTime = 1.0f;
		return true;
	}

	// Calculate the new position on the spline
	glm::vec3 newPosition = getPositionOnSpline(pathTime);
	if (direction != nullptr) {
		front = *direction;
		update();
	}

	// Update camera's position and direction
	position = newPosition;
	return (pathTime >= 1.0f);
}

bool Camera::followPathLookTarget(GLfloat deltaTime, const glm::vec3* targetPosition = nullptr) {
	if (controlPoints.size() < 4) {
		throw std::runtime_error("Fewer than 4 ctr points in followPathLookTarget!");
		return true; // Ensure there are enough control points
	}

	float speed = deltaTime / totalTime;
	pathTime += speed;

	if (pathTime >= 1.0f) {
		pathTime = 1.0f;
		return true;
	}

	// Update camera position along its spline
	glm::vec3 newPosition = getPositionOnSpline(pathTime);
	position = newPosition;

	// Update the camera's orientation to look at the target
	this->targetPosition = *targetPosition;
	front = glm::normalize(*targetPosition - position);

	// Recalculate right and up vectors
	right = glm::normalize(glm::cross(front, worldUp));
	up = glm::normalize(glm::cross(right, front));
	return (pathTime >= 1.0f);
}

glm::vec3 Camera::getTargetPositionOnSpline(float t, const std::vector<glm::vec3>& targetControlPoints) {
	if (targetControlPoints.size() < 4) {
		return glm::vec3(0.0f); // Default position if not enough points
	}

	int numSegments = targetControlPoints.size() - 3;
	int currentSegment = static_cast<int>(t) % numSegments;
	float segmentT = t - static_cast<int>(t);

	glm::vec3 p0 = targetControlPoints[currentSegment];
	glm::vec3 p1 = targetControlPoints[currentSegment + 1];
	glm::vec3 p2 = targetControlPoints[currentSegment + 2];
	glm::vec3 p3 = targetControlPoints[currentSegment + 3];

	return glm::catmullRom(p0, p1, p2, p3, segmentT);
}

Camera::Camera(glm::vec3 startPosition, glm::vec3 startUp, GLfloat startYaw, GLfloat startPitch, GLfloat startMoveSpeed, GLfloat startTurnSpeed)
{
	bool demoMode = false;
	position = startPosition;
	worldUp = startUp;
	yaw = startYaw;
	pitch = startPitch;
	front = glm::vec3(0.0f, 0.0f, -1.0f);

	moveSpeed = startMoveSpeed;
	turnSpeed = startTurnSpeed;

	update();
}

void Camera::keyControl(bool* keys, GLfloat deltaTime)
{
	moveSpeed = 5.0f;
	// calculate velocity based on deltaTime
	GLfloat velocity = moveSpeed * deltaTime;

	// looks into window.keys array and checks if certain key has been pressed 
	if (keys[GLFW_KEY_W])
	{
		position += front * velocity;
	}

	if (keys[GLFW_KEY_S])
	{
		position -= front * velocity;
	}

	if (keys[GLFW_KEY_A])
	{
		position -= right * velocity;
	}

	if (keys[GLFW_KEY_D])
	{
		position += right * velocity;
	}

	if (keys[GLFW_KEY_SPACE])
	{
		position += up * velocity;
	}

	if (keys[GLFW_KEY_LEFT_CONTROL] || keys[GLFW_KEY_RIGHT_CONTROL])
	{
		position -= up * velocity;
	}
}

void Camera::mouseControl(GLfloat xChange, GLfloat yChange)
{
	//applying mouse movement to yaw and pitch for recalculation of camera vectors in update()
	xChange *= turnSpeed;
	yChange *= turnSpeed;

	yaw += xChange;
	pitch += yChange;

	if (pitch > 89.0f)
	{
		pitch = 89.0f;
	}

	if (pitch < -89.0f)
	{
		pitch = -89.0f;
	}

	update();
}

glm::mat4 Camera::calculateViewMatrix()
{
	return glm::lookAt(position, position + front, up);
}

glm::vec3 Camera::getCameraPosition()
{
	return position;
}
glm::vec3 Camera::getCameraDirection()
{
	return glm::normalize(front);
}

void Camera::setPosition(glm::vec3 pos)
{
	position = pos;
}

void Camera::setFront(glm::vec3 fro)
{
	front = fro;
}

void Camera::setTotalTime(GLfloat time)
{
	totalTime = time;
}

void Camera::update()
{
	//recalculate front vector after updating yaw and pitch
	front.x = cos(glm::radians(yaw)) * cos(glm::radians(pitch));
	front.y = sin(glm::radians(pitch));
	front.z = sin(glm::radians(yaw)) * cos(glm::radians(pitch));
	front = glm::normalize(front);

	//calculate right and up vectors from the updated front vector
	right = glm::normalize(glm::cross(front, worldUp));
	up = glm::normalize(glm::cross(right, front));
}


Camera::~Camera()
{
}
