#include "DemoController.h"
#include "Camera.h" 
#include "LinearMovement.h"
#include "CircularMovement.h"
#include "TargetMovement.h"
#include <iostream>
#include <iomanip>
#include <irrKlang.h>
#include <chrono>
#include <ctime>

using namespace irrklang;

DemoController::DemoController(Camera& camera, irrklang::ISound* playingSound) : camera(camera), playingSound(playingSound) {
    demoMode = false;
    cameraPosAtStart = camera.getCameraPosition();
    cameraDirAtStart = camera.getCameraDirection();
    currentMovement = nullptr;
    defineControllPoints();
}

void DemoController::startDemo(Camera& camera) {
    std::cout << "Starting Demo: " << std::endl;
    demoMode = true;

    // Save current camera position and rotation
    cameraPosAtStart = camera.getCameraPosition();
    cameraDirAtStart = camera.getCameraDirection();
    
    // Start the music
    if (playingSound) {
        playingSound->setIsPaused(false);
    }


    // Set variables for the first movement in the queue
    if (!movementQueue.empty()) {
        currentMovement = movementQueue.front();
        camera.setControlPoints(currentMovement->getControlPoints());
        camera.setTotalTime(currentMovement->getDuration());
        startMovementTime = std::chrono::system_clock::now();
        endMovementTime = startMovementTime + std::chrono::milliseconds(static_cast<long long>(currentMovement->getDuration() * 1000));
    }

}

void DemoController::endDemo() {
    std::cout << "Ending Demo: " << std::endl;
    demoMode = false;

    // Stop the Music
    if (playingSound) {
        playingSound->setIsPaused(true);
        playingSound->setPlayPosition(0);
    }
    
    // Set camera back to original position
    camera.setPosition(cameraPosAtStart);
    camera.setFront(cameraDirAtStart);

    // refill the Position Queue for next demo
    defineControllPoints();
    currentMovement = nullptr;
}

void DemoController::processMovementQueue(GLfloat deltaTime) {
    currentTime = std::chrono::system_clock::now();
    // std::cout << "Movement queue size: " << movementQueue.size() << std::endl;
    if (currentMovement != nullptr) {
        if (currentTime >= endMovementTime || currentMovement->finished) {
            
            delete currentMovement;
            movementQueue.pop();

            if (!movementQueue.empty()) {
                // Set up the next movement
                // std::cout << "Next Movement!" << std::endl;
                currentMovement = movementQueue.front();
                camera.setControlPoints(currentMovement->getControlPoints());
                camera.setTotalTime(currentMovement->getDuration());
                startMovementTime = currentTime;
                endMovementTime = startMovementTime + std::chrono::milliseconds(static_cast<long long>(currentMovement->getDuration() * 1000));
            }
            else {
                // No more movements, end Demo
                // std::cout << "Empty Queue!" << std::endl;
                endDemo();
                
                
            }
        }
    }
    else if (currentMovement == nullptr) {
        endDemo();
    }
}

void DemoController::moveCamera(GLfloat deltaTime) {
    
    processMovementQueue(deltaTime);
    if (currentMovement){
        currentMovement->execute(camera, deltaTime);
    }
    
}

void DemoController::printTimePoint(const std::chrono::system_clock::time_point& timePoint) {
    std::time_t tt = std::chrono::system_clock::to_time_t(timePoint);
    std::tm localTime = *std::localtime(&tt);
    auto ms = std::chrono::duration_cast<std::chrono::milliseconds>(timePoint.time_since_epoch()) % 1000; // Get milliseconds

    std::cout << std::put_time(&localTime, "%Y-%m-%d %H:%M:%S") << "." << std::setfill('0') << std::setw(3) << ms.count() << std::endl;
}

void DemoController::defineControllPoints() {
    // Emptie the queue
    std::queue<CameraMovement*> empty;
    std::swap(movementQueue, empty);

    // Defining some useful points
    float cameraMovementRadius = 60.0f;
    glm::vec3 centerCameraPosition = glm::vec3(0.0f, 4.0f, 0.0f);
    glm::vec3 lookDirection = glm::vec3(1.0f, 2.0f, 1.0f);
    glm::vec3 centerCrowd = glm::vec3(-0.111567, 3.05571, 25.8668);
    glm::vec3 djHead = glm::vec3(0.725199, 5.30762, 3.36215);
    glm::vec3 keyboard = glm::vec3(0.0227042, 3.6335, 4.48072);
    
    // Start of the Song (157 BPM) 
    
    // 4x8 Intro I (only robots)
    movementQueue.push(new TargetMovement({
        glm::vec3(-6.88543, 2.34633, 9.82569),
        glm::vec3(-0.900133, 2.34633, 10.2458) ,
        glm::vec3(5.86496, 2.09086, 12.8101) ,
        glm::vec3(10.901, 2.03255, 19.7797) }, beatsToSeconds(8), glm::vec3(-7.76271, 1.91637, 19.4708)));

    movementQueue.push(new LinearMovement({
        glm::vec3(11.0423, 6.80161, 10.5321),
        glm::vec3(7.58718, 5.09735, 14.8525) ,
        glm::vec3(0.82597, 3.43083, 16.8238) ,
        glm::vec3(-6.26628, 2.62692, 13.9253) }, beatsToSeconds(8)));

    movementQueue.push(new TargetMovement({
        glm::vec3(8.76176, 0.826303, 12.9382),
        glm::vec3(2.05547, 0.405972, 13.5131) ,
        glm::vec3(-4.2721, 0.300239, 13.303) ,
        glm::vec3(-9.1891, 0.278344, 13.3637) }, beatsToSeconds(8), glm::vec3(-0.209934, 3.45465, 14.6036)));

    movementQueue.push(new LinearMovement({
        glm::vec3(-12.4745, 3.11573, 9.64776),
        glm::vec3(-7.42263, 2.75987, 12.8662) ,
        glm::vec3(-2.23035, 2.39412, 16.174) ,
        glm::vec3(1.9093, 2.10252, 18.8113) }, beatsToSeconds(8)));
    
    // 4x8 Intro II (DJ & Stage)
    movementQueue.push(new TargetMovement({
        glm::vec3(-6.71936, 3.08097, 5.43983), 
        glm::vec3(-3.30237, 3.06288, 10.088) ,
        glm::vec3(4.65756, 3.06288, 9.07266) ,
        glm::vec3(8.01418, 3.06288, 3.0554) }, beatsToSeconds(8), centerCameraPosition));

    movementQueue.push(new TargetMovement({
        glm::vec3(0.871708, 6.11855, 7.40331),
        glm::vec3(5.14195, 6.3862, 3.02798) ,
        glm::vec3(0.0276613, 5.95835, -0.362814) ,
        glm::vec3(-3.77065, 6.04838, 3.7611) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-2.18009, 3.97627, 4.19685),
        glm::vec3(-0.613278, 4.34403, 4.24693) ,
        glm::vec3(0.692485, 4.18836, 4.28516) ,
        glm::vec3(2.24759, 4.24258, 4.26847) }, beatsToSeconds(8), glm::vec3(0.0497865, 3.59641, 4.31569)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.77201, 6.40015, -0.139998),
        glm::vec3(3.6885, 6.40015, -0.110403) ,
        glm::vec3(-3.81161, 6.40015, -0.0515095) ,
        glm::vec3(-8.2284, 6.40015, -0.0171911) }, beatsToSeconds(8), glm::vec3(-0.75394, 2.76628, 52.6554)));
    
    // 8x8 Beats I (Whole Scene)
    movementQueue.push(new TargetMovement({
        glm::vec3(23.5975, 5.42154, 18.7684),
        glm::vec3(8.91991, 5.42154, 31.8378) ,
        glm::vec3(-9.47911, 5.42154, 32.354) ,
        glm::vec3(-22.4857, 5.42154, 18.8059) }, beatsToSeconds(4), djHead));
    movementQueue.push(new LinearMovement({
        glm::vec3(-0.0278233, 3.35396, 40.6771),
        glm::vec3(-0.125499, 3.82039, 31.2657) ,
        glm::vec3(0.114491, 4.20741, 22.2764) ,
        glm::vec3(-0.0409081, 4.83537, 6.63355) }, beatsToSeconds(4)));


    movementQueue.push(new TargetMovement({
        glm::vec3(-7.7664, 16.74, 11.2885),
        glm::vec3(-8.85948, 12.9987, 26.1867) ,
        glm::vec3(8.40006, 7.23363, 32.4008) ,
        glm::vec3(15.4953, 6.01924, 8.40511) }, beatsToSeconds(8), djHead));

    movementQueue.push(new CircularMovement({
        glm::vec3(0.0764687, 2.65854, 12.2811),
        glm::vec3(-6.28621, 2.41546, 25.7434) ,
        glm::vec3(7.80211, 2.45585, 24.4583) ,
        glm::vec3(0.0764687, 2.65854, 12.2811) }, beatsToSeconds(8), centerCameraPosition));

    movementQueue.push(new TargetMovement({
        glm::vec3(11.6628, 0.920047, 3.57826),
        glm::vec3(11.8085, 2.74642, 3.52088) ,
        glm::vec3(11.9474, 4.48824, 3.46615) ,
        glm::vec3(12.2132, 7.89152, 3.36118) }, beatsToSeconds(8), glm::vec3(-1.87343, 3.05224, 10.4176)));

    movementQueue.push(new TargetMovement({
        glm::vec3(12.0904, 1.73807, 19.1265),
        glm::vec3(5.63206, 1.78905, 20.2382) ,
        glm::vec3(-3.86997, 1.84493, 20.326) ,
        glm::vec3(-12.5048, 1.84493, 17.4658) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(4.96334, 0.159588, 25.9892),
        glm::vec3(11.1372, 0.159588, 21.3323),
        glm::vec3(12.5909, 0.159588, 15.6834),
        glm::vec3(9.35878, 0.318727, 10.1086)}, beatsToSeconds(8), glm::vec3(-0.192702, 2.46498, 17.7186)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.13018, 4.94202, 3.14251),
        glm::vec3(1.58823, 7.88217, 3.26018),
        glm::vec3(-1.47964, 8.14831, 3.08395),
        glm::vec3(-4.75566, 5.47386, 3.00919)}, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-4.97283, 2.36877, 13.2067),
        glm::vec3(-6.72516, 2.36877, 14.9124),
        glm::vec3(-5.21698, 2.36877, 16.9352),
        glm::vec3(-3.51596, 2.35757, 14.7941)}, beatsToSeconds(8), glm::vec3(-6.95515, 2.34009, 14.8137)));

    // 4x8 Beats II
    movementQueue.push(new LinearMovement({
        glm::vec3(-0.766494, 33.1866, 17.5377),
        glm::vec3(-0.768463, 18.7744, 17.1624),
        glm::vec3(-0.438621, 9.98272, 13.4174),
        glm::vec3(0.0305061, 5.62473, 4.81014)}, beatsToSeconds(8))); // top down to dj

    movementQueue.push(new TargetMovement({
        glm::vec3(-1.13244, 4.22949, 2.3736),
        glm::vec3(-2.04092, 4.41303, 5.78475),
        glm::vec3(1.89436, 4.49902, 5.78581),
        glm::vec3(1.6721, 4.80821, 2.00803)}, beatsToSeconds(8), keyboard)); // second keyboard

    movementQueue.push(new TargetMovement({
        glm::vec3(4.77543, 1.24333, 2.63366),
        glm::vec3(4.19425, 1.24333, 8.23989),
        glm::vec3(-4.66799, 1.35615, 7.74363),
        glm::vec3(-6.23293, 1.37051, 2.66316)}, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(2.17546, 2.36683, 9.8213),
        glm::vec3(-1.13139, 2.36683, 13.4683),
        glm::vec3(-3.36904, 2.36683, 15.9361),
        glm::vec3(-6.33153, 2.36683, 19.2034) }, beatsToSeconds(8), glm::vec3(15.6299, 2.28598, 29.461)));

    // 4x8 Beats III (repreating from intro II)
    movementQueue.push(new TargetMovement({
        glm::vec3(-6.71936, 3.08097, 5.43983),
        glm::vec3(-3.30237, 3.06288, 10.088) ,
        glm::vec3(4.65756, 3.06288, 9.07266) ,
        glm::vec3(8.01418, 3.06288, 3.0554) }, beatsToSeconds(8), centerCameraPosition));

    movementQueue.push(new TargetMovement({
        glm::vec3(0.871708, 6.11855, 7.40331),
        glm::vec3(5.14195, 6.3862, 3.02798) ,
        glm::vec3(0.0276613, 5.95835, -0.362814) ,
        glm::vec3(-3.77065, 6.04838, 3.7611) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-2.18009, 3.97627, 4.19685),
        glm::vec3(-0.613278, 4.34403, 4.24693) ,
        glm::vec3(0.692485, 4.18836, 4.28516) ,
        glm::vec3(2.24759, 4.24258, 4.26847) }, beatsToSeconds(8), glm::vec3(0.0497865, 3.59641, 4.31569)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.77201, 6.40015, -0.139998),
        glm::vec3(3.6885, 6.40015, -0.110403) ,
        glm::vec3(-3.81161, 6.40015, -0.0515095) ,
        glm::vec3(-8.2284, 6.40015, -0.0171911) }, beatsToSeconds(8), glm::vec3(-0.75394, 2.76628, 52.6554)));

    // 8x8 Beats IV (repeating from beats I)
    movementQueue.push(new TargetMovement({
        glm::vec3(23.5975, 5.42154, 18.7684),
        glm::vec3(8.91991, 5.42154, 31.8378) ,
        glm::vec3(-9.47911, 5.42154, 32.354) ,
        glm::vec3(-22.4857, 5.42154, 18.8059) }, beatsToSeconds(8), djHead));

    movementQueue.push(new LinearMovement({
        glm::vec3(-0.0278233, 3.35396, 40.6771),
        glm::vec3(-0.125499, 3.82039, 31.2657) ,
        glm::vec3(0.114491, 4.20741, 22.2764) ,
        glm::vec3(-0.0409081, 4.83537, 6.63355) }, beatsToSeconds(8)));

    movementQueue.push(new TargetMovement({
        glm::vec3(-7.7664, 16.74, 11.2885),
        glm::vec3(-8.85948, 12.9987, 26.1867) ,
        glm::vec3(8.40006, 7.23363, 32.4008) ,
        glm::vec3(15.4953, 6.01924, 8.40511) }, beatsToSeconds(8), djHead));

    movementQueue.push(new CircularMovement({
    glm::vec3(0.0764687, 2.65854, 12.2811),
    glm::vec3(-6.28621, 2.41546, 25.7434) ,
    glm::vec3(7.80211, 2.45585, 24.4583) ,
    glm::vec3(0.0764687, 2.65854, 12.2811) }, beatsToSeconds(8), centerCameraPosition));

    movementQueue.push(new TargetMovement({
        glm::vec3(11.6628, 0.920047, 3.57826),
        glm::vec3(11.8085, 2.74642, 3.52088) ,
        glm::vec3(11.9474, 4.48824, 3.46615) ,
        glm::vec3(12.2132, 7.89152, 3.36118) }, beatsToSeconds(8), glm::vec3(-1.87343, 3.05224, 10.4176)));

    movementQueue.push(new TargetMovement({
        glm::vec3(12.0904, 1.73807, 19.1265),
        glm::vec3(5.63206, 1.78905, 20.2382) ,
        glm::vec3(-3.86997, 1.84493, 20.326) ,
        glm::vec3(-12.5048, 1.84493, 17.4658) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(4.96334, 0.159588, 25.9892),
        glm::vec3(11.1372, 0.159588, 21.3323),
        glm::vec3(12.5909, 0.159588, 15.6834),
        glm::vec3(9.35878, 0.318727, 10.1086) }, beatsToSeconds(8), glm::vec3(-0.192702, 2.46498, 17.7186)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.13018, 4.94202, 3.14251),
        glm::vec3(1.58823, 7.88217, 3.26018),
        glm::vec3(-1.47964, 8.14831, 3.08395),
        glm::vec3(-4.75566, 5.47386, 3.00919) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-4.97283, 2.36877, 13.2067),
        glm::vec3(-6.72516, 2.36877, 14.9124),
        glm::vec3(-5.21698, 2.36877, 16.9352),
        glm::vec3(-3.51596, 2.35757, 14.7941) }, beatsToSeconds(8), glm::vec3(-6.95515, 2.34009, 14.8137)));

    // 8x8 Beats V (?)
    movementQueue.push(new TargetMovement({
        glm::vec3(8.76176, 0.826303, 12.9382),
        glm::vec3(2.05547, 0.405972, 13.5131) ,
        glm::vec3(-4.2721, 0.300239, 13.303) ,
        glm::vec3(-9.1891, 0.278344, 13.3637) }, beatsToSeconds(7.9), glm::vec3(-0.209934, 3.45465, 14.6036)));

    movementQueue.push(new LinearMovement({
        glm::vec3(-12.4745, 3.11573, 9.64776),
        glm::vec3(-7.42263, 2.75987, 12.8662) ,
        glm::vec3(-2.23035, 2.39412, 16.174) ,
        glm::vec3(1.9093, 2.10252, 18.8113) }, beatsToSeconds(8)));

    movementQueue.push(new TargetMovement({
        glm::vec3(0.871708, 6.11855, 7.40331),
        glm::vec3(5.14195, 6.3862, 3.02798) ,
        glm::vec3(0.0276613, 5.95835, -0.362814) ,
        glm::vec3(-3.77065, 6.04838, 3.7611) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-2.18009, 3.97627, 4.19685),
        glm::vec3(-0.613278, 4.34403, 4.24693) ,
        glm::vec3(0.692485, 4.18836, 4.28516) ,
        glm::vec3(2.24759, 4.24258, 4.26847) }, beatsToSeconds(8), glm::vec3(0.0497865, 3.59641, 4.31569)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.13018, 4.94202, 3.14251),
        glm::vec3(1.58823, 7.88217, 3.26018),
        glm::vec3(-1.47964, 8.14831, 3.08395),
        glm::vec3(-4.75566, 5.47386, 3.00919) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-4.97283, 2.36877, 13.2067),
        glm::vec3(-6.72516, 2.36877, 14.9124),
        glm::vec3(-5.21698, 2.36877, 16.9352),
        glm::vec3(-3.51596, 2.35757, 14.7941) }, beatsToSeconds(8), glm::vec3(-6.95515, 2.34009, 14.8137)));

    movementQueue.push(new LinearMovement({
        glm::vec3(-0.766494, 33.1866, 17.5377),
        glm::vec3(-0.768463, 18.7744, 17.1624),
        glm::vec3(-0.438621, 9.98272, 13.4174),
        glm::vec3(0.0305061, 5.62473, 4.81014) }, beatsToSeconds(8))); // top down to dj

    movementQueue.push(new TargetMovement({
        glm::vec3(-1.13244, 4.22949, 2.3736),
        glm::vec3(-2.04092, 4.41303, 5.78475),
        glm::vec3(1.89436, 4.49902, 5.78581),
        glm::vec3(1.6721, 4.80821, 2.00803) }, beatsToSeconds(8), keyboard)); // second keyboard

    // 4x8 Beats VI (?)
    movementQueue.push(new TargetMovement({
        glm::vec3(12.0904, 1.73807, 19.1265),
        glm::vec3(5.63206, 1.78905, 20.2382) ,
        glm::vec3(-3.86997, 1.84493, 20.326) ,
        glm::vec3(-12.5048, 1.84493, 17.4658) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(4.96334, 0.159588, 25.9892),
        glm::vec3(11.1372, 0.159588, 21.3323),
        glm::vec3(12.5909, 0.159588, 15.6834),
        glm::vec3(9.35878, 0.318727, 10.1086) }, beatsToSeconds(8), glm::vec3(-0.192702, 2.46498, 17.7186)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.13018, 4.94202, 3.14251),
        glm::vec3(1.58823, 7.88217, 3.26018),
        glm::vec3(-1.47964, 8.14831, 3.08395),
        glm::vec3(-4.75566, 5.47386, 3.00919) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-4.97283, 2.36877, 13.2067),
        glm::vec3(-6.72516, 2.36877, 14.9124),
        glm::vec3(-5.21698, 2.36877, 16.9352),
        glm::vec3(-3.51596, 2.35757, 14.7941) }, beatsToSeconds(8), glm::vec3(-6.95515, 2.34009, 14.8137)));

    // 4x8 Beats VII (?)
    movementQueue.push(new TargetMovement({
        glm::vec3(12.0904, 1.73807, 19.1265),
        glm::vec3(5.63206, 1.78905, 20.2382) ,
        glm::vec3(-3.86997, 1.84493, 20.326) ,
        glm::vec3(-12.5048, 1.84493, 17.4658) }, beatsToSeconds(7.9), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(4.96334, 0.159588, 25.9892),
        glm::vec3(11.1372, 0.159588, 21.3323),
        glm::vec3(12.5909, 0.159588, 15.6834),
        glm::vec3(9.35878, 0.318727, 10.1086) }, beatsToSeconds(8), glm::vec3(-0.192702, 2.46498, 17.7186)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.13018, 4.94202, 3.14251),
        glm::vec3(1.58823, 7.88217, 3.26018),
        glm::vec3(-1.47964, 8.14831, 3.08395),
        glm::vec3(-4.75566, 5.47386, 3.00919) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-4.97283, 2.36877, 13.2067),
        glm::vec3(-6.72516, 2.36877, 14.9124),
        glm::vec3(-5.21698, 2.36877, 16.9352),
        glm::vec3(-3.51596, 2.35757, 14.7941) }, beatsToSeconds(8), glm::vec3(-6.95515, 2.34009, 14.8137)));

    // 1x8 Break? ()    
    movementQueue.push(new TargetMovement({
        glm::vec3(23.5975, 5.42154, 18.7684),
        glm::vec3(8.91991, 5.42154, 31.8378),
        glm::vec3(-9.47911, 5.42154, 32.354),
        glm::vec3(-22.4857, 5.42154, 18.8059) }, beatsToSeconds(8), djHead));
    
    // 7x8 Beats VIII
    movementQueue.push(new TargetMovement({
        glm::vec3(12.0904, 1.73807, 19.1265),
        glm::vec3(5.63206, 1.78905, 20.2382) ,
        glm::vec3(-3.86997, 1.84493, 20.326) ,
        glm::vec3(-12.5048, 1.84493, 17.4658) }, beatsToSeconds(8), djHead));

    movementQueue.push(new LinearMovement({
        glm::vec3(-0.766494, 33.1866, 17.5377),
        glm::vec3(-0.768463, 18.7744, 17.1624),
        glm::vec3(-0.438621, 9.98272, 13.4174),
        glm::vec3(0.0305061, 5.62473, 4.81014) }, beatsToSeconds(8))); // top down to dj

    movementQueue.push(new LinearMovement({
        glm::vec3(-0.0278233, 3.35396, 40.6771),
        glm::vec3(-0.125499, 3.82039, 31.2657) ,
        glm::vec3(0.114491, 4.20741, 22.2764) ,
        glm::vec3(-0.0409081, 4.83537, 6.63355) }, beatsToSeconds(7.9)));

    movementQueue.push(new TargetMovement({
        glm::vec3(-7.7664, 16.74, 11.2885),
        glm::vec3(-8.85948, 12.9987, 26.1867) ,
        glm::vec3(8.40006, 7.23363, 32.4008) ,
        glm::vec3(15.4953, 6.01924, 8.40511) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(4.96334, 0.159588, 25.9892),
        glm::vec3(11.1372, 0.159588, 21.3323),
        glm::vec3(12.5909, 0.159588, 15.6834),
        glm::vec3(9.35878, 0.318727, 10.1086) }, beatsToSeconds(8), glm::vec3(-0.192702, 2.46498, 17.7186)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.13018, 4.94202, 3.14251),
        glm::vec3(1.58823, 7.88217, 3.26018),
        glm::vec3(-1.47964, 8.14831, 3.08395),
        glm::vec3(-4.75566, 5.47386, 3.00919) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-4.97283, 2.36877, 13.2067),
        glm::vec3(-6.72516, 2.36877, 14.9124),
        glm::vec3(-5.21698, 2.36877, 16.9352),
        glm::vec3(-3.51596, 2.35757, 14.7941) }, beatsToSeconds(7.9), glm::vec3(-6.95515, 2.34009, 14.8137)));

    // 4x8 Beats IX (=Intro II)
    movementQueue.push(new TargetMovement({
        glm::vec3(-6.71936, 3.08097, 5.43983),
        glm::vec3(-3.30237, 3.06288, 10.088) ,
        glm::vec3(4.65756, 3.06288, 9.07266) ,
        glm::vec3(8.01418, 3.06288, 3.0554) }, beatsToSeconds(8), centerCameraPosition));

    movementQueue.push(new TargetMovement({
        glm::vec3(0.871708, 6.11855, 7.40331),
        glm::vec3(5.14195, 6.3862, 3.02798) ,
        glm::vec3(0.0276613, 5.95835, -0.362814) ,
        glm::vec3(-3.77065, 6.04838, 3.7611) }, beatsToSeconds(8), djHead));

    movementQueue.push(new TargetMovement({
        glm::vec3(-2.18009, 3.97627, 4.19685),
        glm::vec3(-0.613278, 4.34403, 4.24693) ,
        glm::vec3(0.692485, 4.18836, 4.28516) ,
        glm::vec3(2.24759, 4.24258, 4.26847) }, beatsToSeconds(8), glm::vec3(0.0497865, 3.59641, 4.31569)));

    movementQueue.push(new TargetMovement({
        glm::vec3(6.77201, 6.40015, -0.139998),
        glm::vec3(3.6885, 6.40015, -0.110403) ,
        glm::vec3(-3.81161, 6.40015, -0.0515095) ,
        glm::vec3(-8.2284, 6.40015, -0.0171911) }, beatsToSeconds(8), glm::vec3(-0.75394, 2.76628, 52.6554)));

    // 4x8 Beats X (=Intro I)
    movementQueue.push(new TargetMovement({
        glm::vec3(-6.88543, 2.34633, 9.82569),
        glm::vec3(-0.900133, 2.34633, 10.2458) ,
        glm::vec3(5.86496, 2.09086, 12.8101) ,
        glm::vec3(10.901, 2.03255, 19.7797) }, beatsToSeconds(8), glm::vec3(-7.76271, 1.91637, 19.4708)));

    movementQueue.push(new LinearMovement({
        glm::vec3(11.0423, 6.80161, 10.5321),
        glm::vec3(7.58718, 5.09735, 14.8525) ,
        glm::vec3(0.82597, 3.43083, 16.8238) ,
        glm::vec3(-6.26628, 2.62692, 13.9253) }, beatsToSeconds(8)));

    movementQueue.push(new TargetMovement({
        glm::vec3(8.76176, 0.826303, 12.9382),
        glm::vec3(2.05547, 0.405972, 13.5131),
        glm::vec3(-4.2721, 0.300239, 13.303),
        glm::vec3(-9.1891, 0.278344, 13.3637)}, beatsToSeconds(16), glm::vec3(-0.209934, 3.45465, 14.6036)));

    }

GLfloat DemoController::beatsToSeconds(GLfloat numBeats) {


    GLfloat bpm = 157.0f;
    GLfloat secondsPerBeat = 60 / bpm;
    GLfloat totalSeconds = numBeats * secondsPerBeat;
    return totalSeconds;
}

