// ParticleSystem.cpp

#include "ParticleSystem.h"

ParticleSystem::ParticleSystem(GLuint particleCount, const float TTL, float particleSize, double spawnRatePerSecond, glm::vec3 baseColor, float nearZ, float farZ)
    : particleCount(particleCount), TTL(TTL), particleSize(particleSize), spawnRatePerSecond(spawnRatePerSecond), baseColor(baseColor), nearZ(nearZ), farZ(farZ), movementPattern(nullptr), index(0) {}

ParticleSystem::~ParticleSystem() {}

std::string ParticleSystem::LoadShaderCode(const char* filePath) {
    std::ifstream shaderFile;
    std::stringstream shaderStream;

    shaderFile.open(filePath);
    if (!shaderFile.is_open()) {
        std::cerr << "Failed to open shader file: " << filePath << std::endl;
        return "";
    }

    shaderStream << shaderFile.rdbuf();
    shaderFile.close();

    return shaderStream.str();
}

GLuint ParticleSystem::CompileShader(const char* filePath, GLenum shaderType) {
    std::string shaderCode = LoadShaderCode(filePath);
    const char* shaderSource = shaderCode.c_str();

    GLuint shader = glCreateShader(shaderType);
    glShaderSource(shader, 1, &shaderSource, NULL);
    glCompileShader(shader);

    GLint success;
    GLchar infoLog[512];
    glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
    if (!success) {
        glGetShaderInfoLog(shader, 512, NULL, infoLog);
        std::cerr << "Error compiling shader: " << filePath << "\n" << infoLog << std::endl;
        return 0;
    }

    return shader;
}

GLuint ParticleSystem::LinkProgram(GLuint vertexShader, GLuint geometryShader, GLuint fragmentShader) {
    GLuint program = glCreateProgram();
    glAttachShader(program, vertexShader);
    glAttachShader(program, geometryShader);
    glAttachShader(program, fragmentShader);
    glLinkProgram(program);

    GLint success;
    GLchar infoLog[512];
    glGetProgramiv(program, GL_LINK_STATUS, &success);
    if (!success) {
        glGetProgramInfoLog(program, 512, NULL, infoLog);
        std::cerr << "Error linking program\n" << infoLog << std::endl;
        return 0;
    }

    return program;
}

void ParticleSystem::SetMovementPattern(IParticleMovement* newPattern) {
    movementPattern = newPattern;
}

void ParticleSystem::InitializeParticleBuffers() {
    glGenBuffers(2, posSSBO);
    glGenBuffers(2, velSSBO);

    for (int i = 0; i < 2; ++i) {
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, posSSBO[i]);
        glBufferData(GL_SHADER_STORAGE_BUFFER, particleCount * sizeof(glm::vec4), NULL, GL_DYNAMIC_DRAW);

        glBindBuffer(GL_SHADER_STORAGE_BUFFER, velSSBO[i]);
        glBufferData(GL_SHADER_STORAGE_BUFFER, particleCount * sizeof(glm::vec4), NULL, GL_DYNAMIC_DRAW);
    }

    glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0);
}

void ParticleSystem::AddInitialParticles() {
    std::vector<glm::vec4> positions;
    std::vector<glm::vec4> velocities;

    std::srand(static_cast<unsigned int>(std::time(0)));

    for (GLuint i = 0; i < particleCount; ++i) {
        float x = 0.0f;
        float y = -0.5f;
        float z = 0.0f;
        positions.push_back(glm::vec4(x, y, z, TTL));

        float vx = static_cast<float>(std::rand()) / RAND_MAX * 0.1f - 0.05f;
        float vy = 0.f;
        float vz = static_cast<float>(std::rand()) / RAND_MAX * 0.1f - 0.05f;
        velocities.push_back(glm::vec4(vx, vy, vz, 0.0f));
    }

    glBindBuffer(GL_SHADER_STORAGE_BUFFER, posSSBO[0]);
    glBufferSubData(GL_SHADER_STORAGE_BUFFER, 0, particleCount * sizeof(glm::vec4), positions.data());
    glBindBuffer(GL_SHADER_STORAGE_BUFFER, velSSBO[0]);
    glBufferSubData(GL_SHADER_STORAGE_BUFFER, 0, particleCount * sizeof(glm::vec4), velocities.data());

}

void ParticleSystem::InitializeAtomicCounter() {
    glGenBuffers(1, &atomicCounter);
    glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, atomicCounter);
    glBufferData(GL_ATOMIC_COUNTER_BUFFER, sizeof(GLuint), NULL, GL_DYNAMIC_DRAW);
    GLuint zero = 0;
    glBufferSubData(GL_ATOMIC_COUNTER_BUFFER, 0, sizeof(GLuint), &zero);
    glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, 0);
}

void ParticleSystem::CreateComputeShader() {
    computeShaderProgram = glCreateProgram();
    GLuint computeShader = CompileShader("Shaders/particle_compute.glsl", GL_COMPUTE_SHADER);
    if (!computeShader) return;

    glAttachShader(computeShaderProgram, computeShader);
    glLinkProgram(computeShaderProgram);

    GLint success;
    glGetProgramiv(computeShaderProgram, GL_LINK_STATUS, &success);
    if (!success) {
        GLchar infoLog[1024];
        glGetProgramInfoLog(computeShaderProgram, 1024, nullptr, infoLog);
        std::cerr << "Error linking compute shader program:\n" << infoLog << std::endl;
        glDeleteProgram(computeShaderProgram);
        computeShaderProgram = 0;
    }

    glDeleteShader(computeShader);
}

void ParticleSystem::InitializeParticleRendering() {
    glGenVertexArrays(2, particle_vaos);

    for (int i = 0; i < 2; ++i) {
        glBindVertexArray(particle_vaos[i]);

        glBindBuffer(GL_ARRAY_BUFFER, posSSBO[i]);
        glEnableVertexAttribArray(particle_position_location);
        glVertexAttribPointer(particle_position_location, 4, GL_FLOAT, GL_FALSE, 0, NULL);

        glBindVertexArray(0);
    }

    GLuint vertexShader = CompileShader("Shaders/particle_vertex.glsl", GL_VERTEX_SHADER);
    GLuint geometryShader = CompileShader("Shaders/particle_geometry.glsl", GL_GEOMETRY_SHADER);
    GLuint fragmentShader = CompileShader("Shaders/particle_fragment.glsl", GL_FRAGMENT_SHADER);

    if (!vertexShader || !geometryShader || !fragmentShader) {
        std::cerr << "Error compiling shaders." << std::endl;
        return;
    }

    renderProgram = LinkProgram(vertexShader, geometryShader, fragmentShader);

    glDeleteShader(vertexShader);
    glDeleteShader(geometryShader);
    glDeleteShader(fragmentShader);
}

void ParticleSystem::enableBlendMode() {
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
}

void ParticleSystem::disableBlendMode() {
    glDisable(GL_BLEND);
}

void ParticleSystem::Render(const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix, const glm::vec3& cameraPosition, const GLuint depthTexture, const glm::vec2 screenDimensions) {
    enableBlendMode();
    GLuint textureUnit = 4;
    glUseProgram(renderProgram);

    // Bind the depth texture from G-buffer
    glActiveTexture(GL_TEXTURE0 + textureUnit);
    glBindTexture(GL_TEXTURE_2D, depthTexture);
    glUniform1i(glGetUniformLocation(renderProgram, "depthTexture"), textureUnit);

    // Pass the view and projection matrices
    glm::mat4 modelView = viewMatrix;  // Update this if there's a model transformation
    glUniformMatrix4fv(glGetUniformLocation(renderProgram, "ModelViewMatrix"), 1, GL_FALSE, glm::value_ptr(modelView));
    glUniformMatrix4fv(glGetUniformLocation(renderProgram, "ProjectionMatrix"), 1, GL_FALSE, glm::value_ptr(projectionMatrix));

    glUniform1f(glGetUniformLocation(renderProgram, "zNear"), 0.1f);
    glUniform1f(glGetUniformLocation(renderProgram, "zFar"), 100.0f);
    glUniform2fv(glGetUniformLocation(renderProgram, "screenDimensions"), 1, glm::value_ptr(screenDimensions));

    glUniform1f(glGetUniformLocation(renderProgram, "ParticleSize"), particleSize);
    glUniform3fv(glGetUniformLocation(renderProgram, "BaseColor"), 1, glm::value_ptr(baseColor));

    // Render the particles
    glBindVertexArray(particle_vaos[index]);
    glDrawArrays(GL_POINTS, 0, particleCount);
    glBindVertexArray(0);

    glUseProgram(0);
    disableBlendMode();
}



void ParticleSystem::Initialize() {
    InitializeParticleBuffers();
    InitializeAtomicCounter();
    CreateComputeShader();
    InitializeParticleRendering();
    AddInitialParticles();
}

void ParticleSystem::Update(float deltaTime) {
    static double particles_to_spawn = 0.0;

    glUseProgram(computeShaderProgram);

    static GLuint deltaTLocation = glGetUniformLocation(computeShaderProgram, "DeltaT");
    static GLuint maximumCountLocation = glGetUniformLocation(computeShaderProgram, "MaximumCount");
    static GLuint newParticlePositionLocation = glGetUniformLocation(computeShaderProgram, "NewParticlePosition");
    static GLuint spawnCountLocation = glGetUniformLocation(computeShaderProgram, "SpawnCount");
    


    // Update new particle position using the movement pattern
    if (movementPattern) {
        newParticlePosition = movementPattern->GetPosition(glfwGetTime());
    }
    else {
        std::cerr << "No movement pattern set!" << std::endl;
    }

    glUniform1f(deltaTLocation, deltaTime);
    glUniform1ui(maximumCountLocation, particleCount);
    glUniform3fv(newParticlePositionLocation, 1, glm::value_ptr(newParticlePosition));



    particles_to_spawn += spawnRatePerSecond * deltaTime;
    GLuint spawnCount = 0;
    if (particles_to_spawn >= 1.0) {
        spawnCount = static_cast<GLuint>(particles_to_spawn);
        particles_to_spawn -= spawnCount;
    }
    glUniform1ui(spawnCountLocation, spawnCount);

    // Bind buffers
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, posSSBO[index]);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, velSSBO[index]);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, posSSBO[1 - index]);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, velSSBO[1 - index]);
    glBindBufferBase(GL_ATOMIC_COUNTER_BUFFER, 4, atomicCounter);

    // Reset atomic counter using persistent mapping
    glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, atomicCounter);
    void* atomicPtr = glMapBufferRange(GL_ATOMIC_COUNTER_BUFFER, 0, sizeof(GLuint), GL_MAP_WRITE_BIT | GL_MAP_INVALIDATE_BUFFER_BIT | GL_MAP_UNSYNCHRONIZED_BIT);
    if (atomicPtr) {
        GLuint zero = 0;
        memcpy(atomicPtr, &zero, sizeof(GLuint));
        glUnmapBuffer(GL_ATOMIC_COUNTER_BUFFER);
    }

    // Dispatch compute shader
    GLuint numGroups = (particleCount / 256) + 1;
    glDispatchCompute(numGroups, 1, 1);

    // Synchronize to ensure compute shader has finished
    glMemoryBarrier(GL_ATOMIC_COUNTER_BARRIER_BIT);

    // Read back the atomic counter
    glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, atomicCounter);
    atomicPtr = glMapBufferRange(GL_ATOMIC_COUNTER_BUFFER, 0, sizeof(GLuint), GL_MAP_READ_BIT);
    if (atomicPtr) {
        particleCount = *static_cast<GLuint*>(atomicPtr);
        glUnmapBuffer(GL_ATOMIC_COUNTER_BUFFER);
    }

    // Ensure buffer data is ready for the next stage
    glMemoryBarrier(GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT);

    // Unbind buffers
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, 0);
    glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, 0);
    glBindBufferBase(GL_ATOMIC_COUNTER_BUFFER, 4, 0);

    // Unbind program
    glUseProgram(0);

    // Swap buffers
    index = 1 - index;

    // Debug output
    //printf("ParticleCount: %u\n", particleCount);
}

